/*
 * GDB Stub xfer-spaces support code.
 *
 * Copyright (c) 2012 Alex Rozenman <alex_rozenman@mentor.com>
 * Copyright (c) 2012 Mentor Graphics Corp.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

/*
 * For test use:
 * gcc main.c xfer-spaces.c -o /tmp/test -O3 -g -Wall -pedantic -std=gnu99
 */

#include "xfer-spaces.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#define NAME_SEPARATOR "."
#define INITIAL_OFFSET 0x10

typedef struct DebuggableObject
{
  struct DebuggableObject* next;
  Debuggable* debuggable;
  void* opaque;
  unsigned offset;
} DebuggableObject;

typedef struct DebuggableNode
{
  struct DebuggableNode* sibling;
  struct DebuggableNode* child;
  DebuggableObject* object;
  char* name;
} DebuggableNode;

static DebuggableObject *list, **map;
static DebuggableNode* tree;

static DebuggableObject* alloc_object(Debuggable* debuggable, void* opaque)
{
  DebuggableObject* object =
    (DebuggableObject*)malloc(sizeof(DebuggableObject));
  object->debuggable = debuggable;
  object->opaque = opaque;
  object->offset = 0;
  return object;
}

static DebuggableNode* alloc_node()
{
  DebuggableNode* node =
    (DebuggableNode*)malloc(sizeof(DebuggableNode));
  memset(node, 0, sizeof(DebuggableNode));
  return node;
}

static void dealloc_tree(DebuggableNode* node)
{
  if (node->name)
    free(node->name);
  if (node->sibling)
    dealloc_tree(node->sibling);
  if (node->child)
    dealloc_tree(node->child);
  free(node);
}

static DebuggableNode* find_sibling(DebuggableNode* node, const char* name)
{
  DebuggableNode *ptr;
  for (ptr = node; ptr; ptr = ptr->sibling) {
    if (ptr->name && !strcmp(name, ptr->name)) {
      return ptr;
    }
  }
  return 0;
}

static void merge_record(DebuggableNode* ptr, DebuggableObject* object)
{
  char *saveptr, *token;
  char* path = strdup(object->debuggable->get_name(object->opaque));

  token = strtok_r(path, NAME_SEPARATOR, &saveptr);
  while (token) {
    DebuggableNode *sibling = find_sibling(ptr->child, token);
    if (!sibling) {
      sibling = alloc_node();
      sibling->name = strdup(token);
      sibling->sibling = ptr->child;
      ptr->child = sibling;
      ptr = sibling;
    }
    ptr = sibling;
    token = strtok_r (0, NAME_SEPARATOR, &saveptr);
  }

  free(path);

  ptr->object = object;
}

static void print_indent(FILE* out, unsigned indent, char sym)
{
  unsigned i;
  for (i = 0; i < indent; ++i) {
    fprintf(out, "%c ", sym);
  }
}

void xfer_spaces_declare_object(Debuggable* debuggable, void* opaque)
{
  DebuggableObject* ptr;

  /* Should be before elaboration. */
  assert(!map);

  /* Add to the initial list. */
  ptr = alloc_object(debuggable, opaque);
  ptr->next = list;
  list = ptr;
}

void xfer_spaces_elaborate()
{
  DebuggableObject *ptr;
  unsigned offset = INITIAL_OFFSET;

  /* Create the tree root. */
  tree = alloc_node();
  tree->name = strdup("sc");

  /* Set offsets and merge into the tree. */
  for (ptr = list; ptr; ptr = ptr->next) {
    ptr->offset = offset;
    offset += ptr->debuggable->get_data_size(ptr->opaque);
    merge_record(tree, ptr);
  }

  /* Create the map. */
  map = (DebuggableObject**)malloc(sizeof(*map) * offset);
  memset(map, 0, sizeof(*map));
  for (ptr = list; ptr; ptr = ptr->next) {
    map[ptr->offset] = ptr;
  }
}

static void print_tree(DebuggableNode* node, unsigned indent)
{
  DebuggableNode *ptr;
  for (ptr = node; ptr; ptr = ptr->sibling) {
    print_indent(stdout, indent, '.');
    if (ptr->object) 
      printf("%s \"%s\" [0x%x]\n", ptr->name,
             ptr->object->debuggable->get_name(ptr->object->opaque), ptr->object->offset);
    else
      printf("%s\n", ptr->name);
    print_tree(ptr->child, indent + 1);
  }
}

void xfer_spaces_print_tree()
{
  print_tree(tree, 0);
}

static void dump_xml(FILE* out, DebuggableNode* node, unsigned indent)
{
  DebuggableNode *ptr;
  for (ptr = node; ptr; ptr = ptr->sibling) {
    print_indent(out, indent, ' ');
    if (ptr->object) {
      unsigned data_size = ptr->object->debuggable->get_data_size(ptr->object->opaque);
      fprintf(out, "<reg name=\"%s\" offset=\"%d\" bitsize=\"%d\"/>\n",
              ptr->name, ptr->object->offset, data_size * 8);
    } else {
      fprintf(out, "<group name=\"%s\">\n", ptr->name);
      dump_xml(out, ptr->child, indent + 1);
      print_indent(out, indent, ' ');
      fprintf(out, "</group>\n");
    }
  }
}

void xfer_spaces_dump_xml(FILE* out, const char* annex, const char* conv)
{
  fprintf(out, "<space annex=\"%s\" name=\"%s\">\n", annex, conv);
  dump_xml(out, tree, 0);
  fprintf(out, "</space>\n");
}

unsigned xfer_spaces_read_object(unsigned offset, unsigned len,
                                 unsigned char* data)
{
}

unsigned xfer_spaces_write_object(unsigned offset, unsigned len,
                                  const unsigned char* data)
{
}
